source("dataS/scripts/vehicles/specializations/events/SetTurnedOnEvent.lua");

MixerWagon = {};

function MixerWagon.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Fillable, specializations) and SpecializationUtil.hasSpecialization(TurnOnVehicle, specializations);
end

function MixerWagon:load(savegame)

    self.mixerWagonSuperAllowFillType = self.allowFillType;
    self.allowFillType = Utils.overwrittenFunction(self.allowFillType, MixerWagon.allowFillType);

    self.getCapacity = Utils.overwrittenFunction(self.getCapacity, MixerWagon.getCapacity);
    self.getFillLevel = Utils.overwrittenFunction(self.getFillLevel, MixerWagon.getFillLevel);
    self.setFillLevel = Utils.overwrittenFunction(self.setFillLevel, MixerWagon.setFillLevel);
    self.setUnitFillLevel = Utils.overwrittenFunction(self.setUnitFillLevel, MixerWagon.setUnitFillLevel);
    self.allowUnitFillType = Utils.overwrittenFunction(self.allowUnitFillType, MixerWagon.allowUnitFillType);
    self.getUnitFillType = Utils.overwrittenFunction(self.getUnitFillType, MixerWagon.getUnitFillType);
    self.getCanTipToGround = Utils.overwrittenFunction(self.getCanTipToGround, MixerWagon.getCanTipToGround);

    self.resetFillLevelIfNeeded = Utils.overwrittenFunction(self.resetFillLevelIfNeeded, MixerWagon.resetFillLevelIfNeeded);
    self.resetUnitFillLevelIfNeeded = Utils.overwrittenFunction(self.resetUnitFillLevelIfNeeded, MixerWagon.resetUnitFillLevelIfNeeded);

    self.fillShovelFromGroundValue = Utils.overwrittenFunction(self.fillShovelFromGroundValue, MixerWagon.fillShovelFromGroundValue);
    self.fillShovelFromTrigger = Utils.overwrittenFunction(self.fillShovelFromTrigger, MixerWagon.fillShovelFromTrigger);
    self.mixerWagonBaleTriggerCallback = MixerWagon.mixerWagonBaleTriggerCallback;
    self.getCanBeTurnedOn = Utils.overwrittenFunction(self.getCanBeTurnedOn, MixerWagon.getCanBeTurnedOn);
    self.getIsTurnedOnAllowed = Utils.overwrittenFunction(self.getIsTurnedOnAllowed, MixerWagon.getIsTurnedOnAllowed);
    self.drawMixerWagonHelp = SpecializationUtil.callSpecializationsFunction("drawMixerWagonHelp");

    self.getAllowFillShovel = Utils.overwrittenFunction(self.getAllowFillShovel, MixerWagon.getAllowFillShovel);

    self.mixerWagonNeedsTurnOn = false;
    if self.shovelNodes ~= nil then
        self.mixerWagonNeedsTurnOn = true;
    end

    self.mixerWagonPickupSoundEnabled = false;
    if self.isClient then
        self.sampleMixerWagonPickup = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.mixerWagonPickupSound", nil, self.baseDirectory, self.components[1].node);
        self.sampleMixerWagonStart = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.mixerWagonPickupStartSound", nil, self.baseDirectory);
        self.sampleMixerWagonStop = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.mixerWagonPickupStopSound", nil, self.baseDirectory);
        self.mixerWagonPickupSoundPitchOffset = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.mixerWagonPickupSound#pitchOffset"), 1);
        self.mixerWagonPickupSoundPickupPitchOffset = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.mixerWagonPickupSound#pickupPitchOffset"), 0.8*self.mixerWagonPickupSoundPitchOffset);

        self.mixerWagonMixRotatingParts = Utils.loadRotationNodes(self.xmlFile, {}, "vehicle.mixerWagonRotatingParts.mixerWagonRotatingPart", "mixerWagonMix", self.components);
        self.mixerWagonPickupRotatingParts = Utils.loadRotationNodes(self.xmlFile, {}, "vehicle.mixerWagonRotatingParts.mixerWagonRotatingPart", "mixerWagonPickup", self.components);
        self.mixerWagonPickupScroller = Utils.loadScrollers(self.components, self.xmlFile, "vehicle.mixerWagonRotatingParts.mixerWagonScroller", {}, false);
    end


    local uiScale = g_gameSettings:getValue("uiScale")

    self.mixerWagonFillTypes = {};
    self.fillTypeToMixerWagonFillType = {};
    local i=0;
    while true do
        local baseName = string.format("vehicle.mixerWagonFillTypes.mixerWagonFillType(%d)", i);
        if not hasXMLProperty(self.xmlFile, baseName) then
            break;
        end
        local entry = {};
        entry.fillTypes = {};

        local j=0;
        while true do
            local fillTypeKey = baseName..string.format(".fillType(%d)", j);
            if not hasXMLProperty(self.xmlFile, fillTypeKey) then
                break;
            end
            local fillType = getXMLString(self.xmlFile, fillTypeKey.."#fillType");
            if fillType ~= nil then
                local fillTypeInt = FillUtil.fillTypeNameToInt[fillType];
                if fillTypeInt ~= nil then
                    if self.fillTypeToMixerWagonFillType[fillTypeInt] == nil then
                        entry.fillTypes[fillTypeInt] = true;
                        self.fillTypeToMixerWagonFillType[fillTypeInt] = entry;
                    else
                        print("Warning: Invalid mixerWagonFillType '"..fillType.."' used multiple times in '"..self.configFileName.."'.");
                    end
                else
                    print("Warning: Invalid mixerWagonFillType '"..fillType.."' in '"..self.configFileName.. "'.");
                end
            end

            j = j+1;
        end
        entry.minPercentage = Utils.getNoNil(getXMLFloat(self.xmlFile, baseName.."#minPercentage"), 0)*0.01;
        entry.maxPercentage = Utils.getNoNil(getXMLFloat(self.xmlFile, baseName.."#maxPercentage"), 100)*0.01;
        entry.fillLevel = 0;

        local name = Utils.getNoNil(getXMLString(self.xmlFile, baseName.."#name"), "unknown");
        entry.name = name;
        local i18n = g_i18n;
        if self.customEnvironment ~= nil then
            i18n = _G[self.customEnvironment].g_i18n;
        end
        entry.nameI18N = i18n:getText("fillType_" .. name);

        local fillTypeDesc = FillUtil.fillTypeNameToDesc[name];
        if fillTypeDesc ~= nil then
            local width, height = getNormalizedScreenValues(20*uiScale, 20*uiScale);
            entry.overlay = Overlay:new("MixerWagonFillType", fillTypeDesc.hudOverlayFilenameSmall, 0, 0, width, height)
            entry.overlay:setColor(unpack(g_currentMission.helpBoxTextColor))
        end

        if next(entry.fillTypes) ~= nil then
            entry.statusBarColor = {0.2122, 0.5271, 0.0307, 1};
            entry.statusBarColor2 = {0.2832, 0.0091, 0.0091, 1}
            local width, height = getNormalizedScreenValues(208*uiScale, 8*uiScale);
            entry.statusBar = StatusBar:new(g_baseUIFilename, g_colorBgUVs, nil, {0.0075, 0.0075, 0.0075, 1}, entry.statusBarColor2, nil, 0, 0, width, height);
            table.insert(self.mixerWagonFillTypes, entry);
        end
        i = i+1;
    end


    if self.isServer then
        self.mixerWagonBaleTriggerId = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.mixerWagonBaleTrigger#index"));
        if self.mixerWagonBaleTriggerId ~= nil then
            addTrigger(self.mixerWagonBaleTriggerId, "mixerWagonBaleTriggerCallback", self);
        end
    end

    self.mixingActiveTimerMax = 5000;
    self.mixingActiveTimer = 0;

    self.mixerWagonFillLevelTextOffsetX, self.mixerWagonFillLevelTextOffsetY = getNormalizedScreenValues(28*uiScale, 7*uiScale)
    _, self.mixerWagonHelpHeightPerFruit = getNormalizedScreenValues(0, 22*uiScale)
    _, self.mixerWagonHelpHeightOffset = getNormalizedScreenValues(0, 10*uiScale)

    local width, height = getNormalizedScreenValues(7*uiScale, 12*uiScale)
    self.mixerWagonFillRangeMarkerOverlay = Overlay:new("mixerWagonFillRangeMarkerOverlay", g_baseUIFilename, 0, 0, width, height);
    self.mixerWagonFillRangeMarkerOverlay:setAlignment(Overlay.ALIGN_VERTICAL_MIDDLE, Overlay.ALIGN_HORIZONTAL_LEFT)
    self.mixerWagonFillRangeMarkerOverlay:setUVs(getNormalizedUVs({374, 648, 12, 20}))
    self.mixerWagonFillRangeMarkerOverlay:setColor(unpack(g_currentMission.helpBoxTextColor))

    self.mixerWagonLastPickupTime = -10000;

    self.synchronizeFillLevel = false;

    self.mixerWagon = {};
    self.mixerWagon.fillUnitIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.mixerWagon#fillUnitIndex"), 1);
    self.mixerWagon.unloadInfoIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.mixerWagon#unloadInfoIndex"), 1);
    self.mixerWagon.loadInfoIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.mixerWagon#loadInfoIndex"), 1);
    self.mixerWagon.dischargeInfoIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.mixerWagon#dischargeInfoIndex"), 1);

    self.mixerWagonDirtyFlag = self:getNextDirtyFlag();

    if savegame ~= nil then
        for i=1, table.getn(self.mixerWagonFillTypes) do
            local fillTypeKey = savegame.key..string.format(".mixerWagonFillType%d#fillLevel", i);
            local fillLevel = getXMLFloat(savegame.xmlFile, fillTypeKey);
            if fillLevel == nil then
                fillLevel = 0;
            end
            if fillLevel > 0 then
                fillLevel = fillLevel + self:getFillLevel( next(self.mixerWagonFillTypes[i].fillTypes) );
                self:setUnitFillLevel(self.mixerWagon.fillUnitIndex, fillLevel, next(self.mixerWagonFillTypes[i].fillTypes));
            end
        end
    end

    self.synchronizeFillLevels = false;
end

function MixerWagon:delete()
    if self.isServer then
        if self.mixerWagonBaleTriggerId ~= nil then
            removeTrigger(self.mixerWagonBaleTriggerId);
        end
    end

    for _,entry in pairs(self.mixerWagonFillTypes) do
        if entry.overlay ~= nil then
            entry.overlay:delete()
        end
        entry.statusBar:delete();
    end
    if self.isClient then
        SoundUtil.deleteSample(self.sampleMixerWagonPickup);
        SoundUtil.deleteSample(self.sampleMixerWagonStart);
        SoundUtil.deleteSample(self.sampleMixerWagonStop);
    end

    self.mixerWagonFillRangeMarkerOverlay:delete()
end

function MixerWagon:getSaveAttributesAndNodes(nodeIdent)
    local nodes = "";
    for i=1, table.getn(self.mixerWagonFillTypes) do
        if i>1 then
            nodes = nodes.."\n";
        end
        local fillLevel = self.mixerWagonFillTypes[i].fillLevel;
        nodes = nodes.. nodeIdent..string.format('<mixerWagonFillType%d', i)..' fillLevel="'..fillLevel..'" />';
    end
    return "", nodes;
end

function MixerWagon:readStream(streamId, connection)
    self:setUnitFillLevel(self.mixerWagon.fillUnitIndex, 0, FillUtil.FILLTYPE_UNKNOWN);
    for i=1,table.getn(self.mixerWagonFillTypes) do
        local fillLevel = streamReadFloat32(streamId);
        if fillLevel > 0 then
            local basicFillLevel = self:getUnitFillLevel(self.mixerWagon.fillUnitIndex);
            local mixerWagonFillType = self.mixerWagonFillTypes[i];
            self:setUnitFillLevel(self.mixerWagon.fillUnitIndex, basicFillLevel+fillLevel, next(mixerWagonFillType.fillTypes));
        end
    end
end

function MixerWagon:writeStream(streamId, connection)
    for i=1,table.getn(self.mixerWagonFillTypes) do
        streamWriteFloat32(streamId, self.mixerWagonFillTypes[i].fillLevel);
    end
end

function MixerWagon:readUpdateStream(streamId, timestamp, connection)
    if connection:getIsServer() then
        if streamReadBool(streamId) then
            self.mixingActiveTimer = streamReadUIntN(streamId, 4)/15 * self.mixingActiveTimerMax;

            self:setUnitFillLevel(self.mixerWagon.fillUnitIndex, 0, FillUtil.FILLTYPE_UNKNOWN);
            for i=1,table.getn(self.mixerWagonFillTypes) do
                local fillLevel = streamReadFloat32(streamId);
                if fillLevel > 0 then
                    local basicFillLevel = self:getUnitFillLevel(self.mixerWagon.fillUnitIndex);
                    local mixerWagonFillType = self.mixerWagonFillTypes[i];
                    self:setUnitFillLevel(self.mixerWagon.fillUnitIndex, basicFillLevel+fillLevel, next(mixerWagonFillType.fillTypes));
                end
            end
        end
    end
end

function MixerWagon:writeUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        if streamWriteBool(streamId, bitAND(dirtyMask, self.mixerWagonDirtyFlag) ~= 0) then
            streamWriteUIntN(streamId, math.floor(self.mixingActiveTimer/self.mixingActiveTimerMax*15), 4);
            for i=1,table.getn(self.mixerWagonFillTypes) do
                local percent = 0;
                if self:getCapacity() ~= 0 then
                    percent = Utils.clamp(self.mixerWagonFillTypes[i].fillLevel / self:getUnitCapacity(self.mixerWagon.fillUnitIndex), 0, 1);
                end
                streamWriteFloat32(streamId, self.mixerWagonFillTypes[i].fillLevel);
            end
        end
    end
end

function MixerWagon:mouseEvent(posX, posY, isDown, isUp, button)
end

function MixerWagon:keyEvent(unicode, sym, modifier, isDown)
end

function MixerWagon:update(dt)
    if self.isClient then
        if self.mixingActiveTimer > 0 or self:getIsTurnedOn() or self.tipState == Trailer.TIPSTATE_OPENING or self.tipState == Trailer.TIPSTATE_OPEN then
            if self:getIsActiveForSound() then
                if not SoundUtil.isSamplePlaying(self.sampleMixerWagonStart, 1.8*dt) then
                    SoundUtil.playSample(self.sampleMixerWagonPickup, 0, 0, nil);
                end
            else
                SoundUtil.play3DSample(self.sampleMixerWagonPickup);
            end
            if SoundUtil.isSamplePlaying(self.sampleMixerWagonPickup) then
                if self.mixerWagonLastPickupTime >= g_currentMission.time-200 then
                    SoundUtil.setSamplePitch(self.sampleMixerWagonPickup, self.mixerWagonPickupSoundPickupPitchOffset);
                else
                    SoundUtil.setSamplePitch(self.sampleMixerWagonPickup, self.mixerWagonPickupSoundPitchOffset);
                end
            end
        else
            SoundUtil.stopSample(self.sampleMixerWagonStart, true);
            SoundUtil.stopSample(self.sampleMixerWagonStop, true);
            SoundUtil.stopSample(self.sampleMixerWagonPickup, true);
            SoundUtil.stop3DSample(self.sampleMixerWagonPickup);
        end

        Utils.updateRotationNodes(self, self.mixerWagonPickupRotatingParts, dt, self:getIsActive() and self:getIsTurnedOn());
        Utils.updateScrollers(self.mixerWagonPickupScroller, dt, self:getIsActive() and self:getIsTurnedOn());
        Utils.updateRotationNodes(self, self.mixerWagonMixRotatingParts, dt, self.mixingActiveTimer > 0 or self:getIsTurnedOn() or self.tipState == Trailer.TIPSTATE_OPENING or self.tipState == Trailer.TIPSTATE_OPEN);
    end

    if self.mixingActiveTimer > 0 or self:getIsTurnedOn() or self.tipState == Trailer.TIPSTATE_OPENING or self.tipState == Trailer.TIPSTATE_OPEN then
        self.mixingActiveTimer = self.mixingActiveTimer - dt;
    end
end

function MixerWagon:updateTick(dt)
end

function MixerWagon:draw()
    if self:getIsActiveForInput(true) then
        local numFillTypes = table.getn(self.mixerWagonFillTypes);
        if numFillTypes > 0 then
            local helpHeight = (numFillTypes+1)*self.mixerWagonHelpHeightPerFruit + self.mixerWagonHelpHeightOffset
            g_currentMission:addHelpTextFunction(self.drawMixerWagonHelp, self, helpHeight, nil);
        end
    end
end

function MixerWagon:onDetach()
end

function MixerWagon:onLeave()
end

function MixerWagon:onDeactivateSounds()
    if self.isClient then
        SoundUtil.stopSample(self.sampleMixerWagonPickup, true);
        SoundUtil.stopSample(self.sampleMixerWagonStart, true);
        SoundUtil.stop3DSample(self.sampleMixerWagonPickup);
    end
end

function MixerWagon:resetFillLevelIfNeeded(superFunc, fillType)
    local mixerWagonFillType = self.fillTypeToMixerWagonFillType[fillType];
    if mixerWagonFillType ~= nil then
        return;
    end
    superFunc(self, fillType);
end

function MixerWagon:resetUnitFillLevelIfNeeded(superFunc, fillUnitIndex, fillType)
    if self.fillUnits == nil or self.fillUnits[fillUnitIndex] == nil or self.lastFillLevelChangeTime + 500 > g_currentMission.time then
        return;
    end

    local mixerWagonFillType = self.fillTypeToMixerWagonFillType[fillType];
    if mixerWagonFillType ~= nil then
        return;
    end

    local fillUnit = self.fillUnits[fillUnitIndex];
    if fillUnit.fillTypes[fillType] then
        if fillUnit.currentFillType ~= fillType then
            self:setUnitFillLevel(fillUnitIndex, 0, FillUtil.FILLTYPE_UNKNOWN, true);
        end
    end
end

function MixerWagon:allowFillType(superFunc, fillType, allowEmptying)
    local mixerWagonFillType = self.fillTypeToMixerWagonFillType[fillType];
    if mixerWagonFillType ~= nil then
        return true;
    end
    return superFunc(self, fillType, allowEmptying);
end

function MixerWagon:getCapacity(superFunc, fillType)
    local capacity = 0;
    for _,fillUnit in pairs(self.fillUnits) do
        if fillType == nil or fillType == FillUtil.FILLTYPE_UNKNOWN or fillUnit.fillTypes[fillType] then
            capacity = capacity + fillUnit.capacity;
        end
    end

    if self:getFillLevel() == capacity then
        return self:getFillLevel(fillType);
    end;

    return capacity;
end

function MixerWagon:getFillLevel(superFunc, fillType)
    local totalFillLevel = 0;
    for _, mixerWagonFillType in pairs(self.mixerWagonFillTypes) do
        totalFillLevel = totalFillLevel + mixerWagonFillType.fillLevel;
    end

    return totalFillLevel;
end

function MixerWagon:setFillLevel(superFunc, fillLevel, fillType, force, fillInfo)
    if not self:allowFillType(fillType) then
        return;
    end
    self:setUnitFillLevel(self.mixerWagon.fillUnitIndex, fillLevel, fillType, force, fillInfo)
end

function MixerWagon:setUnitFillLevel(superFunc, fillUnitIndex, fillLevel, fillType, force, fillInfo)
    if fillUnitIndex ~= self.mixerWagon.fillUnitIndex then
        superFunc(self, fillUnitIndex, fillLevel, fillType, force, fillInfo);
        return;
    end

    if not self:allowFillType(fillType, false) then
        return;
    end

    fillLevel = Utils.clamp(fillLevel, 0, self:getUnitCapacity(self.mixerWagon.fillUnitIndex));

    local mixerWagonFillType = self.fillTypeToMixerWagonFillType[fillType];
    if mixerWagonFillType == nil then
        if fillLevel == 0 and fillType == FillUtil.FILLTYPE_UNKNOWN then
            for _,entry in pairs(self.mixerWagonFillTypes) do
                entry.fillLevel = 0;
            end
            superFunc(self, fillUnitIndex, fillLevel, fillType, true, fillInfo);
            return;
        end

        if fillLevel < self:getUnitFillLevel(self.mixerWagon.fillUnitIndex) and self:getUnitFillLevel(self.mixerWagon.fillUnitIndex) > 0 then
            fillLevel = Utils.clamp(fillLevel, 0, self:getUnitCapacity(self.mixerWagon.fillUnitIndex));
            local delta = fillLevel - self:getUnitFillLevel(self.mixerWagon.fillUnitIndex);

            local newFillLevel = 0;
            for _, entry in pairs(self.mixerWagonFillTypes) do
                local entryDelta = delta * (entry.fillLevel / self:getUnitFillLevel(self.mixerWagon.fillUnitIndex));
                entry.fillLevel = math.max(entry.fillLevel + entryDelta, 0);
                newFillLevel = newFillLevel + entry.fillLevel;
            end

            self:raiseDirtyFlags(self.mixerWagonDirtyFlag);

            superFunc(self, fillUnitIndex, newFillLevel, fillType, force, fillInfo);
        else
        end
        return;
    end

    local fillUnitIndex = self.mixerWagon.fillUnitIndex;
    local fillUnit = self.fillUnits[fillUnitIndex];

    local delta = fillLevel - fillUnit.fillLevel;
    local free = fillUnit.capacity - fillUnit.fillLevel;

    if delta > 0 then
        mixerWagonFillType.fillLevel = mixerWagonFillType.fillLevel + math.min(free, delta);
    elseif delta < 0 then
        mixerWagonFillType.fillLevel = math.max(0, mixerWagonFillType.fillLevel + delta);
    end

    if delta > 0 then
        self.mixingActiveTimer = self.mixingActiveTimerMax;
        self.mixerWagonLastPickupTime = g_currentMission.time;
    end

    local newFillLevel = 0;
    for _, mixerWagonFillType in pairs(self.mixerWagonFillTypes) do
        newFillLevel = newFillLevel + mixerWagonFillType.fillLevel;
    end
    newFillLevel = Utils.clamp(newFillLevel, 0, self:getUnitCapacity(fillUnitIndex));

    local newFillType = FillUtil.FILLTYPE_UNKNOWN;

    local isSingleFilled = false;
    for _, mixerWagonFillType in pairs(self.mixerWagonFillTypes) do
        if newFillLevel == mixerWagonFillType.fillLevel then
            isSingleFilled = true;
            newFillType = next(mixerWagonFillType.fillTypes);
            break;
        end
    end

    local isMixingOk = false;
    if not isSingleFilled then
        isMixingOk = true;
        for _, mixerWagonFillType in pairs(self.mixerWagonFillTypes) do
            if mixerWagonFillType.fillLevel < mixerWagonFillType.minPercentage * newFillLevel - 0.01 or mixerWagonFillType.fillLevel > mixerWagonFillType.maxPercentage * newFillLevel + 0.01 then
                isMixingOk = false;
                break;
            end
        end
    end

    if isMixingOk then
        newFillType = FillUtil.FILLTYPE_LIQUIDFERTILIZER;
    elseif not isSingleFilled then
        newFillType = FillUtil.FILLTYPE_FERTILIZERTRASH;
    end

    superFunc(self, fillUnitIndex, newFillLevel, newFillType, true, fillInfo);

    self:raiseDirtyFlags(self.mixerWagonDirtyFlag);

end

function MixerWagon:allowUnitFillType(superFunc, fillUnitIndex, inputFillType, allowEmptying)
    if fillUnitIndex ~= self.mixerWagon.fillUnitIndex then
        return superFunc(self, fillUnitIndex, inputFillType, allowEmptying);
    end

    local isAllowed = false;
    for _, entry in pairs(self.mixerWagonFillTypes) do
        if entry.fillTypes[inputFillType] == true then
            isAllowed = true;
            break;
        end
    end

    if not isAllowed then
        isAllowed = superFunc(self, fillUnitIndex, inputFillType, allowEmptying);
    end

    return isAllowed;
end

function MixerWagon:getUnitFillType(superFunc, fillUnitIndex)

    if fillUnitIndex == self.trailer.fillUnitIndex then
        if self.tipState == Trailer.TIPSTATE_OPENING or self.tipState == Trailer.TIPSTATE_OPEN then

            local currentFillType;

            local fillUnit = self.fillUnits[self.trailer.fillUnitIndex];
            if fillUnit.currentFillType == FillUtil.FILLTYPE_FERTILIZERTRASH and fillUnit.fillLevel > 0 then
                local maxFillType = FillUtil.FILLTYPE_UNKNOWN;
                for _, entry in pairs(self.mixerWagonFillTypes) do
                    if entry.fillLevel > 0 then
                        maxFillType = next(entry.fillTypes);
                        break;
                    end
                end
                currentFillType = maxFillType;
            else
                currentFillType = fillUnit.currentFillType;
            end

            return currentFillType;

        else
            return superFunc(self, fillUnitIndex);
        end
    else
        return superFunc(self, fillUnitIndex);
    end

end

function MixerWagon:getCanTipToGround(superFunc)
    if TipUtil.isValid() then
        local currentFillType;

        local fillUnit = self.fillUnits[self.trailer.fillUnitIndex];
        if fillUnit.currentFillType == FillUtil.FILLTYPE_FERTILIZERTRASH and fillUnit.fillLevel > 0 then
            local maxFillType = FillUtil.FILLTYPE_UNKNOWN;
            for _, entry in pairs(self.mixerWagonFillTypes) do
                if entry.fillLevel > 0 then
                    maxFillType = next(entry.fillTypes);
                    break;
                end
            end
            currentFillType = maxFillType;
        else
            currentFillType = fillUnit.currentFillType;
        end

        if currentFillType == FillUtil.FILLTYPE_UNKNOWN then
            return true;
        end
        return TipUtil.getCanTipToGround(currentFillType);
    else
        return false;
    end
end

function MixerWagon:mixerWagonBaleTriggerCallback(triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
    if onEnter then
        if otherActorId ~= 0 then
            local object = g_currentMission:getNodeObject(otherActorId);
            if object ~= nil then
                if object:isa(Bale) then
                    local fillLevel = object:getFillLevel();
                    local fillType = object:getFillType();
                    if self:allowFillType(fillType, false) then
                        fillLevel = fillLevel + self:getFillLevel();
                        self:setUnitFillLevel(self.mixerWagon.fillUnitIndex, fillLevel, fillType, false, self.fillVolumeLoadInfos[self.mixerWagon.loadInfoIndex]);
                    end
                    object:delete();

                    self.mixingActiveTimer = self.mixingActiveTimerMax;
                    self:raiseDirtyFlags(self.mixerWagonDirtyFlag);
                end
            end
        end
    end
end

function MixerWagon:drawMixerWagonHelp(posY, param)
    if g_currentMission.inGameMessage:getIsVisible() then return; end

    local posX = g_currentMission.helpBoxTextPos1X;
    local posX2 = g_currentMission.helpBoxTextPos2X;

    setTextColor(g_currentMission.helpBoxTextColor[1], g_currentMission.helpBoxTextColor[2], g_currentMission.helpBoxTextColor[3], g_currentMission.helpBoxTextColor[4]);

    posY = posY - self.mixerWagonHelpHeightPerFruit
    setTextBold(true);
    renderText(posX, posY + (self.mixerWagonHelpHeightPerFruit-g_currentMission.helpBoxTextSize)*0.5, g_currentMission.helpBoxTextSize, g_i18n:getText("info_mixingRatio"));
    setTextBold(false);

    local totalFillLevel = 0;
    if self:getUnitFillLevel(self.mixerWagon.fillUnitIndex) > 0 then
        for _, mixerWagonFillType in pairs(self.mixerWagonFillTypes) do
            totalFillLevel = totalFillLevel + mixerWagonFillType.fillLevel;
        end
    end

    for _, mixerWagonFillType in pairs(self.mixerWagonFillTypes) do
        posY = posY - self.mixerWagonHelpHeightPerFruit;

        local percentage = 0;
        if self:getUnitFillLevel(self.mixerWagon.fillUnitIndex) > 0 then
            percentage = mixerWagonFillType.fillLevel / totalFillLevel;
        end

        if mixerWagonFillType.fillLevel > 0 then
            if self:getUnitFillType(self.mixerWagon.fillUnitIndex) ~= FillUtil.FILLTYPE_FERTILIZERTRASH or (percentage >= mixerWagonFillType.minPercentage and percentage <= mixerWagonFillType.maxPercentage) then
                mixerWagonFillType.statusBar:setColor(unpack(mixerWagonFillType.statusBarColor));
            else
                mixerWagonFillType.statusBar:setColor(unpack(mixerWagonFillType.statusBarColor2));
            end
        end
        mixerWagonFillType.statusBar:setPosition(posX2 - mixerWagonFillType.statusBar.width, posY+(self.mixerWagonHelpHeightPerFruit-mixerWagonFillType.statusBar.height)*0.5);
        mixerWagonFillType.statusBar:setValue(percentage);
        mixerWagonFillType.statusBar:render();

        if mixerWagonFillType.overlay ~= nil then
            mixerWagonFillType.overlay:setPosition(posX, posY+(self.mixerWagonHelpHeightPerFruit-mixerWagonFillType.overlay.height)*0.5)
            mixerWagonFillType.overlay:render()
        end

        setTextAlignment(RenderText.ALIGN_LEFT);
        renderText(posX + self.mixerWagonFillLevelTextOffsetX, posY+self.mixerWagonFillLevelTextOffsetY, g_currentMission.helpBoxTextSize, string.format("%1.2f%%", percentage * 100));


        local y = mixerWagonFillType.statusBar.y + mixerWagonFillType.statusBar.height*0.5
        local maxLeft = mixerWagonFillType.statusBar.x + mixerWagonFillType.statusBar.width;
        local xLeft =  math.min(mixerWagonFillType.statusBar.x + mixerWagonFillType.statusBar.width*mixerWagonFillType.minPercentage, maxLeft);
        local xRight = math.min(mixerWagonFillType.statusBar.x + mixerWagonFillType.statusBar.width*mixerWagonFillType.maxPercentage, maxLeft);
        self.mixerWagonFillRangeMarkerOverlay:setInvertX(false)
        self.mixerWagonFillRangeMarkerOverlay:setPosition(xLeft, y)
        self.mixerWagonFillRangeMarkerOverlay:render()
        self.mixerWagonFillRangeMarkerOverlay:setInvertX(true)
        self.mixerWagonFillRangeMarkerOverlay:setPosition(xRight, y)
        self.mixerWagonFillRangeMarkerOverlay:render()
   end
end

function MixerWagon:onTurnedOn(noEventSend)
    if self.isClient then
        SoundUtil.stopSample(self.sampleMixerWagonStop, true);
        SoundUtil.stopSample(self.sampleMixerWagonPickup, true);
        SoundUtil.stop3DSample(self.sampleMixerWagonPickup);
        if self:getIsActiveForSound() then
            SoundUtil.playSample(self.sampleMixerWagonStart, 1, 0, nil);
        end
    end
end

function MixerWagon:onTurnedOff(noEventSend)
    if self.isClient then
        SoundUtil.stopSample(self.sampleMixerWagonStart, true);
        SoundUtil.stopSample(self.sampleMixerWagonPickup, true);
        SoundUtil.stop3DSample(self.sampleMixerWagonPickup);
        if self:getIsActiveForSound() then
            SoundUtil.playSample(self.sampleMixerWagonStop, 1, 0, nil);
        end
    end
end

function MixerWagon:getCanBeTurnedOn(superFunc)
    if superFunc ~= nil then
        return superFunc(self);
    end
    return true;
end

function MixerWagon:getIsTurnedOnAllowed(superFunc, isTurnedOn)
    if superFunc ~= nil then
        return superFunc(self, isTurnedOn);
    end
    return true;
end

function MixerWagon:getAllowFillShovel(superFunc, fillType)
    if self:getIsTurnedOn() then
        if superFunc ~= nil then
            return superFunc(self, fillType);
        end
    end
    return false;
end

function MixerWagon:fillShovelFromGroundValue(superFunc, areas, fillType)
    local value = false;
    if superFunc ~= nil then
        value = superFunc(self, areas, fillType)
    end
    if value then
        self.mixerWagonLastPickupTime = g_currentMission.time;
    end
    return value;
end

function MixerWagon:fillShovelFromTrigger(superFunc, shovelTrigger, deltaFillLevel, fillType, dt)
    local value = 0;
    if superFunc ~= nil then
        value = superFunc(self, shovelTrigger, deltaFillLevel, fillType, dt)
    end
    if value > 0 then
        self.mixerWagonLastPickupTime = g_currentMission.time;
    end
    return value;
end
